/* Copyright (c) 2002, Marek Michalkiewicz
   Copyright (c) 2007 Joerg Wunsch
   Copyright (c) 2008 Carlos Lamas
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.

   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. */

#ifndef _UTIL_DELAY_BASIC_H_
#define _UTIL_DELAY_BASIC_H_ 1

#include <stdint.h>

/** \file */
/** \defgroup util_delay_basic <util/delay_basic.h>: Basic busy-wait delay loops
    \code
    #include <util/delay_basic.h>
    \endcode

    The functions in this header file implement simple delay loops
    that perform a busy-waiting.  They are typically used to
    facilitate short delays in the program execution.  They are
    implemented as count-down loops with a well-known CPU cycle
    count per loop iteration.  As such, no other processing can
    occur simultaneously.  It should be kept in mind that the
    functions described here do not disable interrupts.

    In general, for long delays, the use of hardware timers is
    much preferrable, as they free the CPU, and allow for
    concurrent processing of other events while the timer is
    running.  However, in particular for very short delays, the
    overhead of setting up a hardware timer is too much compared
    to the overall delay time.

    Four inline functions are provided for the actual delay algorithms
    and several macros. The macros are intended to be used only with
    constant arguments (integer or floating point). The compiler
    performs all the arithmetics and inserts the required delay loop
    and/or padding instructions.

    Variables with the qualifier 'const' are not recognized as constant
    values at compile time; therefore, they are not valid to be used as
    arguments to the macros.

*/

#if !defined(__DOXYGEN__)
static __inline__ void _delay_loop_1(uint8_t __count)  __attribute__((always_inline));
static __inline__ void _delay_loop_2(uint16_t __count) __attribute__((always_inline));
static __inline__ void _delay_loop_3(uint32_t __count) __attribute__((always_inline));
static __inline__ void _delay_loop_4(uint32_t __count) __attribute__((always_inline));
#endif /* !defined(__DOXYGEN__) */


/** \ingroup util_delay_basic

    Delay loop using an 8-bit counter \c __count, so up to 256
    iterations are possible. (The value 256 would have to be passed
    as 0.)  The loop executes three CPU cycles per iteration, not
    including the overhead the compiler needs to setup the counter
    register.

    Thus, at a CPU speed of 1 MHz, delays of up to 768 microseconds
    can be achieved.

    \note Last iteration takes two CPU cycles only.
*/
void
_delay_loop_1(uint8_t __count)
{
    __asm__ volatile (
        "1: dec %0" "\n\t"
        "brne 1b"
        : "=r" (__count)
        : "0" (__count)
    );
}


/** \ingroup util_delay_basic

    Delay loop using a 16-bit counter \c __count, so up to 65536
    iterations are possible. (The value 65536 would have to be
    passed as 0.)  The loop executes four CPU cycles per iteration,
    not including the overhead the compiler requires to setup the
    counter register pair.

    Thus, at a CPU speed of 1 MHz, delays of up to about 262.1
    milliseconds can be achieved.

    \note Last iteration takes three CPU cycles only.
 */
void
_delay_loop_2(uint16_t __count)
{
    __asm__ volatile (
        "1: sbiw %0,1" "\n\t"
        "brne 1b"
        : "=w" (__count)
        : "0" (__count)
    );
}


/** \ingroup util_delay_basic

    Delay loop using a 24-bit counter \c __count, so up to 16,777,216
    iterations are possible. (The value 16,777,216 would have to be
    passed as 0.)  The loop executes five CPU cycles per iteration,
    not including the overhead the compiler requires to setup the
    counter registers. The most significant byte in the 32 bit
    parameter passed is ignored.

    Thus, at a CPU speed of 1 MHz, delays of up to about 83.88
    seconds can be achieved.

    \note Last iteration takes four CPU cycles only.
 */
void
_delay_loop_3(uint32_t __count)
{
    __asm__ volatile (

            "1:  subi    %A0, 1"    "\n\t"
            "    sbci    %B0, 0"    "\n\t"
            "    sbci    %C0, 0"    "\n\t"
            "    brne    1b"
        :
            "=d" (__count)
        :
            "0" (__count)
    );
}


/** \ingroup util_delay_basic

    Delay loop using a 32-bit counter \c __count, so up to 4,294,967,296
    iterations are possible. (The value 4,294,967,296 would have to be
    passed as 0.)  The loop executes six CPU cycles per iteration,
    not including the overhead the compiler requires to setup the
    counter registers.

    Thus, at a CPU speed of 1 MHz, delays of up to about 7.15 hours
    can be achieved.

    \note Last iteration takes five CPU cycles only.
 */
void
_delay_loop_4(uint32_t __count)
{
    __asm__ volatile (

            "1:  subi    %A0, 1"    "\n\t"
            "    sbci    %B0, 0"    "\n\t"
            "    sbci    %C0, 0"    "\n\t"
            "    sbci    %D0, 0"    "\n\t"
            "    brne    1b"
        :
            "=d" (__count)
        :
            "0" (__count)
    );
}


#if !defined(__DOXYGEN__)

/* Internal use macros */

#define __roundc(__cc) ((double)(__cc) + 0.5)


#if defined(__OPTIMIZE__)


#define _delay_loop_const_1(__count)  _delay_loop_1(__count)

#define _delay_loop_const_2(__count)  _delay_loop_2(__count)

#define _delay_loop_const_3(__count)                                  \
    do {                                                              \
        uint16_t __count_lo = (uint16_t)(uint32_t)(__count);          \
        uint8_t __count_hi = (uint8_t)((uint32_t)(__count) >> 16);    \
        __asm__ __volatile__(                                         \
                "1:  sbiw    %0, 1"            "\n\t"                 \
                "    sbc     %1, __zero_reg__" "\n\t"                 \
                "    brne    1b"                                      \
            :                                                         \
                "=w" (__count_lo),                                    \
                "=r" (__count_hi)                                     \
            :                                                         \
                "0" (__count_lo),                                     \
                "1" (__count_hi)                                      \
        );                                                            \
    } while (0)

#define _delay_loop_const_4(__count)                                  \
    do {                                                              \
        uint16_t __count_lo = (uint16_t)(uint32_t)(__count);          \
        uint16_t __count_hi = (uint16_t)((uint32_t)(__count) >> 16);  \
        __asm__ __volatile__(                                         \
                "1:  sbiw    %0, 1"             "\n\t"                \
                "    sbc     %A1, __zero_reg__" "\n\t"                \
                "    sbc     %B1, __zero_reg__" "\n\t"                \
                "    brne    1b"                                      \
            :                                                         \
                "=w" (__count_lo),                                    \
                "=r" (__count_hi)                                     \
            :                                                         \
                "0" (__count_lo),                                     \
                "1" (__count_hi)                                      \
        );                                                            \
    } while (0)

/*
   Loop sizes table (optimization ENABLED & constant argument)

                          iteration    whole loop     compiler
                          (n times)    correction     overhead
   _delay_loop_const_1        3            -1            +1
   _delay_loop_const_2        4            -1            +2
   _delay_loop_const_3        5            -1            +3
   _delay_loop_const_4        6            -1            +4
*/

/* Short version (loop counter rounded) */

#define __ticksS1(__cc)   (((double)(__cc) + 1. - 1.) / 3. + 0.5)
#define __ticksS2(__cc)   (((double)(__cc) + 1. - 2.) / 4. + 0.5)
#define __ticksS3(__cc)   (((double)(__cc) + 1. - 3.) / 5. + 0.5)
#define __ticksS4(__cc)   (((double)(__cc) + 1. - 4.) / 6. + 0.5)

/* Exact version (CPU cycle counter rounded) */

#define __ticksE1(__cc)   (((double)(__cc) + 0.5 + 1. - 1.) / 3.)
#define __ticksE2(__cc)   (((double)(__cc) + 0.5 + 1. - 2.) / 4.)
#define __ticksE3(__cc)   (((double)(__cc) + 0.5 + 1. - 3.) / 5.)
#define __ticksE4(__cc)   (((double)(__cc) + 0.5 + 1. - 4.) / 6.)

#define __deltaE1(__cc)   ((double)(__cc) - ((uint64_t)__ticksE1(__cc) * 3. - 1. + 1.) + 0.5)
#define __deltaE2(__cc)   ((double)(__cc) - ((uint64_t)__ticksE2(__cc) * 4. - 1. + 2.) + 0.5)
#define __deltaE3(__cc)   ((double)(__cc) - ((uint64_t)__ticksE3(__cc) * 5. - 1. + 3.) + 0.5)
#define __deltaE4(__cc)   ((double)(__cc) - ((uint64_t)__ticksE4(__cc) * 6. - 1. + 4.) + 0.5)


#else  /* !defined(__OPTIMIZE__) */


#define _delay_loop_const_1(__count)                        \
    do {                                                    \
        __asm__ __volatile__(                               \
                "    mov     __tmp_reg__, r24"    "\n\t"    \
                "    ldi     r24, lo8(%0)"        "\n\t"    \
                "1:  dec     r24"                 "\n\t"    \
                "    brne    1b"                  "\n\t"    \
                "    mov     r24, __tmp_reg__"              \
            :                                               \
            :                                               \
                "i" (__count)                               \
        );                                                  \
    } while (0)

#define _delay_loop_const_2(__count)                        \
    do {                                                    \
        __asm__ __volatile__(                               \
                "    movw    __tmp_reg__, r24"    "\n\t"    \
                "    ldi     r24, lo8(%0)"        "\n\t"    \
                "    ldi     r25, hi8(%0)"        "\n\t"    \
                "1:  sbiw    r24, 1"              "\n\t"    \
                "    brne    1b"                  "\n\t"    \
                "    movw    r24, __tmp_reg__"    "\n\t"    \
                "    clr     __zero_reg__"                  \
            :                                               \
            :                                               \
                "i" (__count)                               \
        );                                                  \
    } while (0)

#define _delay_loop_const_3(__count)                               \
    do {                                                           \
        uint16_t __count16;                                        \
        __asm__ __volatile__(                                      \
                "    ldi     %A0, hlo8(%1)"              "\n\t"    \
                "    mov     __tmp_reg__, %A0"           "\n\t"    \
                "    ldi     %A0, lo8(%1)"               "\n\t"    \
                "    ldi     %B0, hi8(%1)"               "\n\t"    \
                "1:  sbiw    %0, 1"                      "\n\t"    \
                "    sbc     __tmp_reg__, __zero_reg__"  "\n\t"    \
                "    brne    1b"                                   \
            :                                                      \
                "=&w" (__count16)                                  \
            :                                                      \
                "i" (__count)                                      \
        );                                                         \
    } while (0)

#define _delay_loop_const_4(__count)                     \
    do {                                                 \
        uint32_t __count32;                              \
        __asm__ __volatile__(                            \
                "    ldi     %A0, lo8(%1)"     "\n\t"    \
                "    ldi     %B0, hi8(%1)"     "\n\t"    \
                "    ldi     %C0, hlo8(%1)"    "\n\t"    \
                "    ldi     %D0, hhi8(%1)"    "\n\t"    \
                "1:  subi    %A0, 1"           "\n\t"    \
                "    sbci    %B0, 0"           "\n\t"    \
                "    sbci    %C0, 0"           "\n\t"    \
                "    sbci    %D0, 0"           "\n\t"    \
                "    brne    1b"                         \
            :                                            \
                "=&d" (__count32)                        \
            :                                            \
                "i" (__count)                            \
        );                                               \
    } while (0)


/*
   Loop size table (optimization DISABLED & constant argument)

                          iteration    whole loop     compiler
                          (n times)    correction     overhead
   _delay_loop_const_1        3            +2             0
   _delay_loop_const_2        4            +4             0
   _delay_loop_const_3        5            +3            +4
   _delay_loop_const_4        6            +3            +8
*/

/* Short version */

#define __ticksS1(__cc)   (((double)(__cc) - 2.     ) / 3. + 0.5)
#define __ticksS2(__cc)   (((double)(__cc) - 4.     ) / 4. + 0.5)
#define __ticksS3(__cc)   (((double)(__cc) - 3. - 4.) / 5. + 0.5)
#define __ticksS4(__cc)   (((double)(__cc) - 3. - 8.) / 6. + 0.5)

/* Exact version (CPU cycle counter rounded) */

#define __ticksE1(__cc)   (((double)(__cc) + 0.5 - 2.     ) / 3.)
#define __ticksE2(__cc)   (((double)(__cc) + 0.5 - 4.     ) / 4.)
#define __ticksE3(__cc)   (((double)(__cc) + 0.5 - 3. - 4.) / 5.)
#define __ticksE4(__cc)   (((double)(__cc) + 0.5 - 3. - 8.) / 6.)

#define __deltaE1(__cc)   ((double)(__cc) - ((uint64_t)__ticksE1(__cc) * 3. + 2.     ) + 0.5)
#define __deltaE2(__cc)   ((double)(__cc) - ((uint64_t)__ticksE2(__cc) * 4. + 4.     ) + 0.5)
#define __deltaE3(__cc)   ((double)(__cc) - ((uint64_t)__ticksE3(__cc) * 5. + 3. + 4.) + 0.5)
#define __deltaE4(__cc)   ((double)(__cc) - ((uint64_t)__ticksE4(__cc) * 6. + 3. + 8.) + 0.5)


#endif /* !defined(__OPTIMIZE__) */


#define __WAIT_1_CYCLE()  __asm__ __volatile__ ("nop")

#define __WAIT_2_CYCLES() __asm__ __volatile__ ("rjmp .")

#define __WAIT_3_CYCLES() __asm__ __volatile__ ("rjmp ."    "\n\t"    \
                                                "nop")

#define __WAIT_4_CYCLES() __asm__ __volatile__ ("rjmp ."    "\n\t"    \
                                                "rjmp .")

#define __WAIT_5_CYCLES() __asm__ __volatile__ ("rjmp ."    "\n\t"    \
                                                "rjmp ."    "\n\t"    \
                                                "nop")

#define __WAIT_6_CYCLES() __asm__ __volatile__ ("rjmp ."    "\n\t"    \
                                                "rjmp ."    "\n\t"    \
                                                "rjmp .")


#define __SHORT_DELAY(__cc)          \
    do {                             \
        if ((__cc) >= 6) {           \
            /* 6 wait states */      \
            __WAIT_6_CYCLES();       \
        } else if ((__cc) == 5) {    \
            /* 5 wait states */      \
            __WAIT_5_CYCLES();       \
        } else if ((__cc) == 4) {    \
            /* 4 wait states */      \
            __WAIT_4_CYCLES();       \
        } else if ((__cc) == 3) {    \
            /* 3 wait states */      \
            __WAIT_3_CYCLES();       \
        } else if ((__cc) == 2) {    \
            /* 2 wait states */      \
            __WAIT_2_CYCLES();       \
        } else if ((__cc) == 1) {    \
            /* 1 wait state */       \
            __WAIT_1_CYCLE();        \
        } else {                     \
            /* do nothing */         \
        }                            \
    } while (0)

#endif /* !defined(__DOXYGEN__) */


/** \ingroup util_delay_basic

    \def _delay_cycles(__cycles)

    Delay loop using a CPU cycle counter value \c __cycles. This
    macro calls the appropriate delay loop function or inserts simple
    delay instructions to obtain a delay aproximated to the given
    value.

    The parameter has to be an integer constant or a floating point
    constant known at compile time. The smallest delay loop fitting
    the delay is inserted. For very short delays (less than 7 CPU
    cycles) the code generated does not include any loop at all, just
    a few instructions for sequential execution.

    The delays are generally not accurate to the CPU cycle
    (_delay_exact_cycles() should be used for this). Instead, the loop
    counter is rounded to the nearest value. The loop counters are
    also adjusted to compensate the overhead introduced by the
    compiler pre and post loop. The precision varies with the duration
    of a single step in the loop delay used (+/- 1.5 cycles for
    _delay_loop1(), +/- 2 cycles for _delay_loop2(), +/- 2.5 cycles
    for _delay_loop3() and +/- 3 cycles for _delay_loop4()). Cycle
    counter values under 7 should be accurate to the CPU cycle
    precision (i.e. +/- 0.5 cycles).

    Optimization is not required. The compiler will not generate
    floating point code at all in any optimization level. If the
    optimization is disabled and the stack is placed in external
    memory the delay loops 3 & 4, with compiler overhead, will
    take a few CPU cycles more than expected.
 */
#define _delay_cycles(__cycles)                                              \
    do {                                                                     \
        /* prevent non constant parameter passing */                         \
        __asm__ __volatile__ ( "" : : "F" ((double)(__cycles)) );            \
        if (__ticksS4(__cycles) >= 4294967296.0) {                           \
            /* maximum time 32 bit counter loop */                           \
            _delay_loop_const_4(0);                                          \
        } else if (__ticksS3(__cycles) >= 16777217.0) {                      \
            /* 32 bit counter loop */                                        \
            _delay_loop_const_4(          (uint32_t)__ticksS4(__cycles));    \
        } else if (__ticksS2(__cycles) >= 65537.0) {                         \
            /* 24 bit counter loop */                                        \
            _delay_loop_const_3(          (uint32_t)__ticksS3(__cycles));    \
        } else if (__ticksS1(__cycles) >= 257.0) {                           \
            /* 16 bit counter loop */                                        \
            _delay_loop_const_2((uint16_t)(uint32_t)__ticksS2(__cycles));    \
        } else if (__roundc(__cycles) >= 7.0) {                              \
            /* 8 bit counter loop */                                         \
            _delay_loop_const_1( (uint8_t)(uint16_t)__ticksS1(__cycles));    \
        } else if (__roundc(__cycles) >= 1.0) {                              \
            /* Insert a short delay for less than 7 cycles */                \
            __SHORT_DELAY((uint8_t)__roundc(__cycles));                      \
        } else {                                                             \
            /* zero or negative => do nothing */                             \
        }                                                                    \
    } while (0)


/** \ingroup util_delay_basic

    \def _delay_exact_cycles(__cycles)

    Accurate delay loop using a CPU cycle counter value \c __cycles.
    This macro expands to the appropriate delay loop function and
    complementary padding instructions to obtain the nearest delay
    to the argument entered. This accuracy is payed with code size
    increase; maximum (per delay) 1 instruction for 8 bit loop, 2
    instructions for 16 bit and 24 bit loops and 3 instructions for
    32 bit loop.

    The parameter has to be an integer or floating point constant
    known at compile time. The macro acts as a wrapper to insert
    a delay choosing the smallest loop type that fits the delay.

    The delays are accurate to the CPU cycle. The loop counters are
    adjusted to compensate the overhead introduced by the compiler
    for the loop management. Padding instructions are inserted after
    the loops. The delays should be exact to the CPU cycle precision
    (i.e. +/- 0.5 cycles). Big values can be afected by rounding
    errors due to current reduced floating point arithmetics
    precision (the mantissa is 24 bit only).

    Optimization is not required. The compiler will not generate
    floating point code at all in any optimization level. If the
    optimization is disabled and the stack is placed in external
    memory the delay loops 3 & 4, with compiler overhead, will
    take a few CPU cycles more than expected.
 */
#define _delay_exact_cycles(__cycles)                                        \
    do {                                                                     \
        /* prevent non constant parameter passing */                         \
        __asm__ __volatile__ ( "" : : "F" ((double)(__cycles)) );            \
        if (__ticksE4(__cycles) >= 4294967296.0) {                           \
            /* maximum time 32 bit counter loop */                           \
            _delay_loop_const_4(0);                                          \
        } else if (__ticksE3(__cycles) >= 16777217.0) {                      \
            /* 32 bit counter loop */                                        \
            _delay_loop_const_4(          (uint32_t)__ticksE4(__cycles));    \
            __SHORT_DELAY((uint8_t)__deltaE4(__cycles));                     \
        } else if (__ticksE2(__cycles) >= 65537.0) {                         \
            /* 24 bit counter loop */                                        \
            _delay_loop_const_3(          (uint32_t)__ticksE3(__cycles));    \
            __SHORT_DELAY((uint8_t)__deltaE3(__cycles));                     \
        } else if (__ticksE1(__cycles) >= 257.0) {                           \
            /* 16 bit counter loop */                                        \
            _delay_loop_const_2((uint16_t)(uint32_t)__ticksE2(__cycles));    \
            __SHORT_DELAY((uint8_t)__deltaE2(__cycles));                     \
        } else if (__roundc(__cycles) >= 7.0) {                              \
            /* 8 bit counter loop */                                         \
            _delay_loop_const_1( (uint8_t)(uint16_t)__ticksE1(__cycles));    \
            __SHORT_DELAY((uint8_t)__deltaE1(__cycles));                     \
        } else if (__roundc(__cycles) >= 1.0) {                              \
            /* Insert a short delay for less than 7 cycles */                \
            __SHORT_DELAY((uint8_t)__roundc(__cycles));                      \
        } else {                                                             \
            /* zero or negative => do nothing */                             \
        }                                                                    \
    } while (0)


/** \ingroup util_delay_basic

    \def _delay_f_us(__freq, __us)

    Delay microseconds at given frequency. Processor master clock
    frequency has to be passed as a parameter.

    \c __freq is the CPU frequency in Hz. \c __us is the time to delay
    in microseconds. Both values must be integer or floating point
    constant values known at compile time or the compiler will emit an
    error.

    The maximal possible delay is 25.7698e9 us / __freq in MHz.
 */
#define _delay_f_us(__freq, __us)    \
    _delay_cycles((__us) * (double)(__freq) / 1e6)


/** \ingroup util_delay_basic

    \def _delay_f_ms(__freq, __ms)

    Delay milliseconds at given frequency. Processor master clock
    frequency has to be passed as a parameter.

    \c __freq is the CPU frequency in Hz. \c __ms is the time to delay
    in milliseconds. Both values must be integer or floating point
    constant values known at compile time or the compiler will emit an
    error.

    The maximal possible delay is 25.7698e6 ms / __freq in MHz.
 */
#define _delay_f_ms(__freq, __ms)    \
    _delay_cycles((__ms) * (double)(__freq) / 1e3)


/** \ingroup util_delay_basic

    \def _delay_f_s(__freq, __s)

    Delay seconds at given frequency. Processor master clock frequency
    has to be passed as a parameter.

    \c __freq is the CPU frequency in Hz. \c __s is the time to delay
    in seconds. Both values must be integer or floating point constant
    values known at compile time or the compiler will emit an error.

    The maximal possible delay is 25.7698e3 s / __freq in MHz.
 */
#define _delay_f_s(__freq, __s)    \
    _delay_cycles((__s) * (double)(__freq))


/** \ingroup util_delay_basic

    \def _delay_exact_f_us(__freq, __us)

    Accurate delay in microseconds at given frequency. Processor
    master clock frequency has to be passed as a parameter.

    \c __freq is the CPU frequency in Hz. \c __us is the time to delay
    in microseconds. Both values must be integer or floating point
    constant values known at compile time or the compiler will emit an
    error.

    The maximal possible delay is 25.7698e9 us / __freq in MHz.
 */
#define _delay_exact_f_us(__freq, __us)    \
    _delay_exact_cycles((__us) * (double)(__freq) / 1e6)


/** \ingroup util_delay_basic

    \def _delay_exact_f_ms(__freq, __ms)

    Accurate delay in milliseconds at given frequency. Processor
    master clock frequency has to be passed as a parameter.

    \c __freq is the CPU frequency in Hz. \c __ms is the time to delay
    in milliseconds. Both values must be integer or floating point
    constant values known at compile time or the compiler will emit an
    error.

    The maximal possible delay is 25.7698e6 ms / __freq in MHz.
 */
#define _delay_exact_f_ms(__freq, __ms)    \
    _delay_exact_cycles((__ms) * (double)(__freq) / 1e3)


/** \ingroup util_delay_basic

    \def _delay_exact_f_s(__freq, __s)

    Accurate delay in seconds at given frequency. Processor master
    clock frequency has to be passed as a parameter.

    \c __freq is the CPU frequency in Hz. \c __s is the time to delay
    in seconds. Both values must be integer or floating point constant
    values known at compile time or the compiler will emit an error.

    The maximal possible delay is 25.7698e3 s / __freq in MHz.
 */
#define _delay_exact_f_s(__freq, __s)    \
    _delay_exact_cycles((__s) * (double)(__freq))


#endif /* _UTIL_DELAY_BASIC_H_ */
